"""List-like classes

Contents:
    Listing (base.Bunch, MutableSequence): drop-in replacement for a python
        `list` with additional functionality.
    DictList (Listing): iterable with both `dict` and `list` interfaces. Stored
        items must be hashable or have a `name` attribute.

To Do:


"""
from __future__ import annotations

import copy
import dataclasses
from collections.abc import Hashable, Mapping, MutableSequence
from typing import TYPE_CHECKING, Any

from . import base, utilities

if TYPE_CHECKING:
    from .kinds import GenericList


@dataclasses.dataclass
class Listing(base.Bunch, MutableSequence):
    """Basic `list` replacement.

    A `Listing` differs from an ordinary python `list` in ways required by
    inheriting from `Bunch`: `add`, `delete`, and `subset` methods, and allowing
    the "+" operator to join `Listing`s with other `list`-like objects) and in 1
    other way:
        1) It includes a `prepend` method for adding one or more items to the
            beginning of the stored list.

    The `add` method attempts to extend `contents` with the item to be added.
    If this fails, it appends the item to `contents`.

    Args:
        contents: items to store in a `list`. Defaults to an empty `list`.

    """

    contents: GenericList = dataclasses.field(default_factory = list)

    """ Instance Methods """

    def add(self, item: Any | GenericList) -> None:
        """Tries to extend `contents` with `item`. Otherwise, it appends.

        The method will extend all passed sequences, except str types, which it
        will append.

        Args:
            item: item(s) to add to `contents`.

        """
        if utilities._is_sequence(item = item):
            self.contents.extend(item)
        else:
            self.contents.append(item)
        return

    def delete(self, item: int) -> None:
        """Deletes item at the index in `contents`.

        Args:
            item: index in `contents` to delete.

        """
        del self.contents[item]
        return

    def insert(self, index: int, item: Any) -> None:
        """Inserts `item` at `index` in `contents`.

        Args:
            index: index to insert `item` at.
            item: object to be inserted.

        """
        self.contents.insert(index, item)
        return

    def prepend(self, item: Any | GenericList) -> None:
        """Prepends `item` to `contents`.

        If `item` is a non-str sequence, `prepend` adds its contents to the
        stored list in the order they appear in `item`.

        Args:
            item: item(s) to prepend to `contents`.

        """
        if utilities._is_sequence(item = item):
            for thing in reversed(item):
                self.prepend(item = thing)
        else:
            self.insert(0, item)
        return

    def subset(
        self,
        include: Any | GenericList | None = None,
        exclude: Any | GenericList | None = None) -> Listing:
        """Returns a new instance with a subset of `contents`.

        This method applies `include` before `exclude` if both are passed. If
        `include` is None, all existing items will be added to the new subset
        class instance before `exclude` is applied.

        Args:
            include (Optional[Any | GenericList]): item(s) to include in
                the new instance. Defaults to None.
            exclude (Optional[Any | GenericList]): item(s) to exclude in
                the new instance. Defaults to None.

        Raises:
            ValueError: if `include` and `exclude` are both None.

        Returns:
            Listing: with only items from `include` and no items in `exclude`.

        """
        if include is None and exclude is None:
            raise ValueError('include or exclude must not be None')
        if include is None:
            contents = copy.deepcopy(self.contents)
        else:
            include = list(utilities._iterify(include))
            contents = [i for i in self.contents if i in include]
        if exclude is not None:
            exclude = list(utilities._iterify(exclude))
            contents = [i for i in contents if i not in exclude]
        new_listing = copy.deepcopy(self)
        new_listing.contents = contents
        return new_listing

    """ Dunder Methods """

    def __getitem__(self, index: Any) -> Any:
        """Returns value(s) for `key` in `contents`.

        Args:
            index (Any): index to search for in `contents`.

        Returns:
            Any: item stored in `contents` at key.

        """
        return self.contents[index]

    def __setitem__(self, index: Any, value: Any) -> None:
        """Sets `key` in `contents` to `value`.

        Args:
            index (Any): index to set `value` to in `contents`.
            value (Any): value to be set at `key` in `contents`.

        """
        self.contents[index] = value
        return


@dataclasses.dataclass
class DictList(Listing):
    """Iterable that has both a dict and list interfaces.

    DictList combines the functionality and interfaces of python dicts and lists.
    It allows duplicate keys and list-like iteration while supporting the easier
    access methods of dictionaries. In order to support this hybrid approach to
    iterables, DictList can only store items that are hashable or have a `name`
    attribute or property that contains or returns a hashable value.

    A DictList inherits the differences between a Listing and an ordinary python
    list.

    A DictList differs from a Listing in 3 significant ways:
        1) It only stores hashable items or objects for which a str name can be
            derived (using the namify function).
        2) DictList has an interface of both a dict and a list, but stores a list.
            DictList does this by taking advantage of the `name` attribute or
            hashability of stored items. A `name` or hash acts as a key to
            create the facade of a dict with the items in the stored list
            serving as values. This allows for duplicate keys for storing items,
            simpler iteration than a dict, and support for returning multiple
            matching items. This design comes at the expense of lookup speed. As
            a result, DictList should only be used if a high volume of access
            calls is not anticipated. Ordinarily, the loss of lookup speed
            should have negligible effect on overall performance.
        3) DictLists should not store int types. This ensures that when, for
            example, a `dictlist_instance[3]` is called, the item at that index is
            returned. If int types are stored, that call would create
            uncertainty as to whether an index or item should be returned. By
            design, int types are assumed to be calls to return the item at that
            index.
        4) When using dict access methods, a list of matches may be returned
            because a DictList allows duplicate pseudo-keys to be used.

    Args:
        contents (MutableSequence[Hashable]): items to store that are hashable
            or have a `name` attribute. Defaults to an empty list.
        default_factory (Optional[Any]): default value to return or default
            function to call when the `get` method is used. Defaults to None.

    """

    contents: MutableSequence[Hashable] = dataclasses.field(
        default_factory = list)
    default_factory: Any | None = None

    """ Instance Methods """

    def delete(self, item: Any | int) -> None:
        """Deletes item in `contents`.

        If `item` is not an int type, this method looks for a matching `name`
        attribute in the stored instances and deletes all such items. If `key`
        is an int type, only the item at that index is deleted.

        Args:
            item (Any, int): name or index in `contents` to delete.

        """
        if isinstance(item, int):
            del self.contents[item]
        else:
            self.contents = [
                c for c in self.contents if utilities._namify(c) != item]
        return

    def get(self, key: Hashable, default: Any | None = None) -> Any:
        """Returns value in `contents` or default options.

        Args:
            key (Hashable): key for value in `contents`.
            default (Optional[Any]): default value to return if `key` is not
                found in `contents`.

        Raises:
            KeyError: if `key` is not in the DictList and `default` and the
                `default_factory` attribute are both None.

        Returns:
            Any: value matching key in `contents` or `default_factory` value.

        """
        try:
            return self[key]
        except (KeyError, TypeError) as error:
            if default is not None:
                return default
            if self.default_factory is None:
                raise KeyError(f'{key} is not in the DictList') from error
            try:
                return self.default_factory()
            except TypeError:
                return self.default_factory

    def items(self) -> tuple[tuple[Hashable, ...], tuple[Any, ...]]:
        """Emulates python dict `items` method.

        Returns:
            tuple[tuple[Hashable, ...], tuple[Any, ...]]: a tuple equivalent to
                dict.items(). A DictList cannot actually create an ItemsView
                because that would eliminate any duplicate keys, which are
                permitted by DictList.

        """
        return tuple(zip(self.keys(), self.values(), strict = True))

    def keys(self) -> tuple[Hashable, ...]:
        """Emulates python dict `keys` method.

        Returns:
            tuple[Hashable, ...]: a tuple equivalent to dict.keys(). A DictList
                cannot actually create an KeysView because that would eliminate
                any duplicate keys, which are permitted by DictList.

        """
        return tuple(utilities._namify(c) for c in self.contents)

    def setdefault(self, value: Any) -> None:
        """Sets default value to return when `get` method is used.

        Args:
            value (Any): default value to return.

        """
        self.default_factory = value
        return

    def update(self, items: Mapping[Any, Any]) -> None:
        """Mimics the dict `update` method by extending `contents` with `items`.

        Args:
            items (Mapping[Any, Any]): items to add to the `contents` attribute.
                The values of `items` are added to `contents` and the keys
                become the `name` attributes of those values. As a result, the
                keys of `items` are discarded. To mimic `dict.update`, the
                passed `items` values are added to `contents` by the `extend`
                method which adds the values to the end of `contents`.

        """
        self.extend(list(items.values()))
        return

    def values(self) -> tuple[Any, ...]:
        """Emulates python dict `values` method.

        Returns:
            tuple[Any, ...]: a tuple equivalent to dict.values(). A DictList
                cannot actually create an ValuesView because that would
                eliminate any duplicate keys, which are permitted by DictList.

        """
        return tuple(self.contents)

    """ Dunder Methods """

    def __getitem__(self, key: Hashable | int) -> Any:
        """Returns value(s) for `key` in `contents`.

        If `key` is not an int type, this method looks for a matching `name`
        attribute in the stored instances.

        If `key` is an int type, this method returns the stored item at the
        corresponding index.

        If only one match is found, a single item is returned. If more are
        found, a `DictList` or `DictList` subclass with the matching `name`
        attributes is returned.

        Args:
            key (Hashable, int): name of an item or index to search for
                in `contents`.

        Returns:
            Any: value(s) stored in `contents` that correspond to `key`. If
                there is more than one match, the return is a DictList or DictList
                subclass with that matching stored items.

        """
        if isinstance(key, int):
            return self.contents[key]
        matches = [
            c for c in self.contents if utilities._namify(c) == key]
        if not matches:
            raise KeyError(f'{key} is not in {self.__class__.__name__}')
        if len(matches) == 1:
            return matches[0]
        else:
            return matches

    def __setitem__(self, key: Any | int, value: Any) -> None:
        """Sets `key` in `contents` to `value`.

        Args:
            key (Any | int): if key isn`t an int, it is ignored (since the
                `name` attribute of the value will be acting as the key). In
                such a case, the `value` is added to the end of `contents`. If
                key is an int, `value` is assigned at the that index number in
                `contents`.
            value (Any): value to be paired with `key` in `contents`.

        """
        if isinstance(key, int):
            self.contents[key] = value
        else:
            self.add(value)
        return



"""Base class for extensible, flexible, lightweight collection types.

Contents:
    Bunch (Collection, abc.ABC): base class for collections in `bunches`. It
        requires subclasses to have `add`, `delete`, and `subset` methods.

To Do:


"""
from __future__ import annotations

import abc
import dataclasses
from collections.abc import Collection, Hashable, Iterator
from typing import TYPE_CHECKING, Any, Self

from . import settings

if TYPE_CHECKING:
    from .kinds import SubsetReturns


@dataclasses.dataclass
class Bunch(Collection, abc.ABC):
    """Base for general `bunches` collections.

    A Bunch differs from a general python Collection in 5 ways:
        1) It must include an `add` method which provides the default mechanism
            for adding new items to the collection. `add` allows a subclass to
            designate the preferred method of adding to the collections`s stored
            data without replacing other access methods.
        2) It must include a `delete` method which provides the default
            mechanism for deleting items in the collection. `delete` is called
            by the `__delitem__` dunder method to delete stored items.
        3) A subclass must include a `subset` method with optional `include` and
            `exclude` parameters for returning a subset of the Bunch subclass.
        4) It supports the '+' operator being used to join a Bunch subclass
            instance of the same python type (mapping, sequence, tuple, etc.).
            The '+' operator calls the Bunch subclass `add` method to implement
            how the added item(s) is/are added to the Bunch subclass instance.
        5) It offers an accessible `contents` attribute that contains the native
            Python type for the `Bunch` subclass. This allows easy reversion to
            the simple type or simple bypassing of the `Bunch` subclass methods.

    Args:
        contents: stored collection of items.

    """

    contents: Collection[Any]

    """ Required Subclass Methods """

    @abc.abstractmethod
    def add(self, item: Any, *args: Any, **kwargs: Any) -> None:
        """Adds `item` to `contents`.

        Args:
            item: item to add to `contents`.
            args: positional arguments.
            kwargs: keyword arguments.

        """

    @abc.abstractmethod
    def delete(self, item: Any, *args: Any, **kwargs: Any) -> None:
        """Deletes `item` from `contents`.

        Args:
            item: item or key to delete in `contents`.
            args: positional arguments.
            kwargs: keyword arguments.

        Raises:
            KeyError: if `item` is not in `contents`. Subclasses should
                implement this error.

        """

    @abc.abstractmethod
    def subset(
        self,
        include: Collection[Any] | Any | None = None,
        exclude: Collection[Any] | Any | None = None,
        returns: SubsetReturns = settings._SUBSET_RETURN) -> Bunch:
        """Returns a new instance with a subset of `contents`.

        This method applies `include` before `exclude` if both are passed. If
        `include` is None, all existing items will be added to the new subset
        class instance before `exclude` is applied.

        Args:
            include: item(s) to include in the new `Bunch`. Defaults to `None`.
            exclude: item(s) to exclude from the new `Bunch`. Defaults to
                `None`.
            returns: whether to return a new instance of the `Bunch` subclass
                ('class'), a deep copy of the subclass instance ('copy') or the
                simple native Python type ('simple'). Defaults to the global
                setting stored in `settings._SUBSET_RETURN`.

        """

    """ Dunder Methods """

    def __add__(self, other: Any) -> Self:
        """Combines argument with `contents` using the `add` method.

        Args:
            other: item to add to `contents` using the `add` method.

        """
        self.add(item = other)
        return self

    def __iadd__(self, other: Any) -> Self:
        """Combines argument with `contents` using the `add` method.

        Args:
            other: item to add to `contents` using the `add` method.

        """
        self.add(item = other)
        return self

    def __delitem__(self, item: Hashable) -> Self:
        """Deletes `item` from `contents`.

        Args:
            item: item or key to delete in `contents`.

        Raises:
            KeyError: if `item` is not in `contents`.

        """
        self.delete(item = item)
        return self

    def __iter__(self) -> Iterator[Any]:
        """Returns iterator of `contents`.

        Returns:
            Iterator: of `contents`.

        """
        return iter(self.contents)

    def __len__(self) -> int:
        """Returns length of `contents`.

        Returns:
            int: length of `contents`.

        """
        return len(self.contents)



"""Settings for `bunches`.

Contents:


To Do:


"""
from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from . import utilities

if TYPE_CHECKING:
    from .kinds import SubsetReturns


_ALL_KEYS: list[Any] = ['all', 'All', ['all'], ['All']]
_DEFAULT_KEYS: list[Any] = [
    'default', 'defaults', 'Default', 'Defaults', ['default'], ['defaults'],
    ['Default'], ['Defaults']]
_KEY_NAMER: Callable[[object | type[Any]], str] = utilities._namify
_METHOD_NAMER: Callable[[object | type[Any]], str] = (
    lambda x: f'from_{utilities._namify(x)}')
_NONE_KEYS: list[Any] = ['none', 'None', ['none'], ['None']]
_SUBSET_RETURN: SubsetReturns = 'class'


def set_key_namer(namer: Callable[[object | type[Any]], str]) -> None:
    """Sets the global default function used to name items.

    Args:
        namer: function that returns a str name of any item passed.

    Raises:
        TypeError: if 'namer' is not callable.

    """
    if isinstance(namer, Callable):
        globals()['KEYER'] = namer
    else:
        raise TypeError('namer argument must be a callable')

def set_method_namer(namer: Callable[[object | type[Any]], str]) -> None:
    """Sets the global default function used to name factory methods.

    Args:
        namer (Callable[[object | Type[Any]], str]): function that returns a
            str name of any item passed.

    Raises:
        TypeError: if 'namer' is not callable.

    """
    if isinstance(namer, Callable):
        globals()['_METHOD_NAMER'] = namer
    else:
        raise TypeError('namer argument must be a callable')



"""Tests for core classes

To Do:

"""
from __future__ import annotations
import dataclasses

import bunches


@dataclasses.dataclass
class TestClass(object):

    name: str = 'something'


@dataclasses.dataclass
class AnotherClass(object):

    name: str = 'another'


@dataclasses.dataclass
class ThirdClass(object):

    name: str = 'third'


# def test_proxy():
#     wrapped = TestClass()
#     proxy = bunches.Proxy(contents = wrapped)
#     proxy.id = 4543
#     assert proxy.name == 'something'
#     assert proxy.id == 4543
#     assert hasattr(proxy, 'id')
#     del proxy.id
#     assert not hasattr(proxy, 'id')
#     return

def test_listing():
    listing = bunches.Listing(contents = ['a', 'b', 'c'])
    assert listing[1] == 'b'
    listing.add(item = 'd')
    assert listing[3] == 'd'
    listing.insert(2, 'zebra')
    assert listing[2] == 'zebra'
    listing.append('e')
    assert listing[5] == 'e'
    listing.extend(['f', 'g'])
    assert listing[7] == 'g'
    sub_listing = listing.subset(
        include = ['a', 'b', 'c', 'd', 'zebra'],
        exclude = 'd')
    assert sub_listing.contents == ['a', 'b', 'zebra', 'c']
    sub_listing.remove('c')
    assert sub_listing.contents == ['a', 'b', 'zebra']
    listing.clear()
    return

def test_hybrid():
    hybrid = bunches.DictList(contents = ['a', 'b', 'c'])
    hybrid.setdefault(value = 'No')
    assert hybrid.get('tree') == 'No'
    assert hybrid[1] == 'b'
    hybrid.add(item = 'd')
    assert hybrid[3] == 'd'
    hybrid.insert(2, 'zebra')
    assert hybrid[2] == 'zebra'
    sub_hybrid = hybrid.subset(
        include = ['a', 'b', 'c', 'd', 'zebra'],
        exclude = 'd')
    assert sub_hybrid.contents == ['a', 'b', 'zebra', 'c']
    sub_hybrid.remove('c')
    assert sub_hybrid.contents == ['a', 'b', 'zebra']
    assert hybrid[0] == 'a'
    assert hybrid['zebra'] == 'zebra'
    hybrid.append('b')
    assert hybrid.values() == tuple(['a', 'b', 'zebra', 'c', 'd', 'b'])
    assert hybrid.keys() == tuple(['a', 'b', 'zebra', 'c', 'd', 'b'])
    hybrid.remove('b')
    assert hybrid.contents == ['a', 'zebra', 'c', 'd', 'b']
    hybrid.clear()
    test_class = TestClass()
    hybrid.add(test_class)
    assert hybrid.keys() == tuple(['something'])
    assert hybrid.values() == tuple([test_class])
    return

def test_dictionary():
    alt_created = bunches.Dictionary.fromkeys(
        keys = ['a', 'b', 'c'],
        value = 'tree')
    assert alt_created['a'] == 'tree'
    dictionary = bunches.Dictionary(
        contents = {'a': 'b', 'c': 'd'},
        default_factory = 'Nada')
    assert dictionary.get('f') == 'Nada'
    assert dictionary['a'] == 'b'
    dictionary.add({'e': 'f'})
    assert dictionary['e'] == 'f'
    subset = dictionary.subset(include = ['a', 'e'])
    assert 'a' in subset.keys()
    assert 'b' not in subset.keys()
    assert 'a' not in subset.values()
    assert 'b' in subset.values()
    return

def test_catalog():
    catalog = bunches.Catalog(contents = {'tester': TestClass})
    catalog.add({'another': AnotherClass})
    catalog.add({'a_third': ThirdClass()})
    assert 'tester' in catalog
    assert 'another' in catalog
    assert 'a_third' in catalog
    assert catalog['all'] == catalog['default']
    assert catalog['None'] == None
    catalog.default = 'tester'
    assert catalog['default'] == catalog['tester']
    del catalog[['tester', 'another']]
    assert 'tester' not in catalog
    assert len(catalog) == 1
    return

def test_repository():
    repository = bunches.Repository()
    repository.add(AnotherClass())
    repository.add(ThirdClass, 'random_name')
    assert 'another' in repository
    assert 'random_name' in repository
    repository.delete('random_name')
    assert 'random_name' not in repository
    return

if __name__ == '__main__':
    # test_proxy()
    test_listing()
    test_hybrid()
    test_dictionary()
    test_catalog()
    test_repository()




"""Types and type aliases for the package.

Contents:


To Do:


"""
from __future__ import annotations

import dataclasses
import sys
from collections.abc import Hashable, MutableMapping, MutableSequence
from collections.abc import Set as AbstractSet
from typing import Any, Literal, TypeAlias

if sys.version_info < (3, 12):
    GenericDict: TypeAlias = MutableMapping[Hashable, Any]
    GenericList: TypeAlias = MutableSequence[Any]
    GenericSet: TypeAlias = AbstractSet[Any]
    SubsetReturns: TypeAlias = Literal['class', 'copy', 'simple']
else:
    type GenericDict = MutableMapping[Hashable, Any]
    type GenericList = MutableSequence[Any]
    type GenericSet = AbstractSet[Any]
    type SubsetReturns = Literal['class', 'copy', 'simple']


@dataclasses.dataclass
class _MISSING_VALUE:  # noqa: N801
    """Sentinel object for a missing data or parameter.

    This follows the same pattern as the `_MISSING_TYPE` class in the builtin
    dataclasses library.
    https://github.com/python/cpython/blob/3.10/Lib/dataclasses.py#L182-L186

    Because None is sometimes a valid argument or data option, this class
    provides an alternative that does not create the confusion that a default of
    None can sometimes lead to.

    """

    pass  # noqa: PIE790


# _MISSING, instance of MISSING_VALUE, should be used for missing values as an
# alternative to None. This provides a fuller repr and traceback.
_MISSING = _MISSING_VALUE()



"""System and functions for inferring object and class names.

Contents:


To Do:


"""
from __future__ import annotations

import copy
import inspect
import re
from collections.abc import Collection, Iterable, Sequence
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .kinds import GenericDict, SubsetReturns


def _capitalify(item: str) -> str:
    """Converts a snake case `str` to capital case.

    Args:
        item: `str` to convert.

    Returns:
        'item' converted to capital case.

    """
    return item.replace('_', ' ').title().replace(' ', '')

def _is_sequence(item: Any) -> bool:
    """Returns if 'item' is a sequence but not a `str`.

    Args:
        item: object to examine.

    Returns:
        If 'item' is a sequence but not a `str`.

    """
    if not inspect.isclass(item):
        item = item.__class__
    return issubclass(item, Sequence) and not issubclass(item, str)

def _iterify(item: Any) -> Iterable:
    """Returns `item` as an iterable, but does not iterate `str` types.

    Args:
        item: item to turn into an iterable.

    Returns:
        Iterable of `item`. A `str` type will be stored as a single item in an
            iterable wrapper.

    """
    if item is None:
        return iter(())
    elif isinstance(item, str | bytes):
        return iter([item])
    else:
        try:
            return iter(item)
        except TypeError:
            return iter((item,))

def _namify(item: Any, /, default: str | None = None) -> str | None:
    """Returns `str` name representation of 'item'.

    Args:
        item: item to determine a `str` name.
        default: default name to return if other methods at name creation fail.

    Returns:
        str: a name representation of 'item.'

    """
    if isinstance(item, str):
        return item
    elif (
        hasattr(item, 'name')
        and not inspect.isclass(item)
        and isinstance(item.name, str)):
        return item.name
    else:
        try:
            return _snakify(item.__name__)
        except AttributeError:
            if item.__class__.__name__ is not None:
                return _snakify(item.__class__.__name__)
            else:
                return default

def _return_subset(
    subset: Collection,
    existing: Collection,
    returns: SubsetReturns) -> Collection:
    """Returns a subset of an item.

    Args:
        subset: native Python subset of data from a `Collection`.
        existing: a subclasss instance of `Collection`.
        returns: the type to be be returned by the function.

    Returns:
        A `Collection` with a `subset` of data.

    """
    if returns == "class":
        return existing.__class__(subset)
    elif returns == "copy":
        new_collection = copy.deepcopy(existing)
        new_collection.contents = subset
        return new_collection
    elif returns == "simple":
        return subset
    else:
        message = 'returns argument must be "class", "copy", or "simple"'
        raise ValueError(message)

def _snakify(item: str) -> str:
    """Converts a capitalized `str` to snake case.

    Args:
        item: `str`/. to convert.

    Returns:
        str: 'item' converted to snake case.

    """
    item = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', item)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', item).lower()

def _uniquify(
    key: str,
    dictionary: GenericDict,
    index: int | None = 1) -> str:
    """Creates a unique key name to avoid overwriting an item in 'dictionary'.

    The function is 1-indexed so that the first attempt to avoid a duplicate
    will be: "old_name2".

    Args:
        key: name of key to test.
        dictionary: `dict` for which a unique key name is sought.
        index: current index number for suffix. Defaults to 1.

    Returns:
        str: unique key name for 'dictionary'.

    """
    if key not in dictionary:
        return key
    counter = index
    while True:
        counter += 1
        if counter > 2:  # noqa: PLR2004
            key = key.removesuffix(str(counter - 1))
        key = ''.join([key, str(counter)])
        if key not in dictionary:
            return key



"""Drop-in replacements for Python collections"""

from __future__ import annotations

__version__ = '0.1.3'

__author__: str = 'Corey Rayburn Yung'


from .base import Bunch
from .mappings import Catalog, ChainDict, Dictionary, Repository
from .sequences import DictList, Listing

__all__: list[str] = [
    "Bunch",
    "Catalog",
    "ChainDict",
    "Dictionary",
    "DictList",
    "Listing",
    "Repository"]



"""Extensible, flexible, lightweight dict-like classes

Contents:
    Dictionary (base.Bunch, MutableMapping): drop-in replacement for a python
        dict with some added functionality.
    Catalog (Dictionary): wildcard-accepting dict which is primarily intended
        for storing different options and strategies. It also returns lists of
        matches if a list of keys is provided.
    ChainDict (Dictionary): combines the additional functionality of
        Dictionary with collections.ChainMap from the Python builtin library.
    Repository (Dictionary): stores items using inferred keys when the `add`
        method is called. This is useful for mappings where the key naming is
        controlled by the mapping instead of based on the passed arguments.
        Support for guaranteed unique key creation (using an integer counter) is
        provided out of the box based on the `overwrite` argument.

To Do:


"""
from __future__ import annotations

import contextlib
import copy
import dataclasses
import itertools
from collections.abc import (
    Collection,
    Hashable,
    MutableMapping,
    MutableSequence,
    Sequence,
)
from typing import TYPE_CHECKING, Any

from . import base, settings, utilities

if TYPE_CHECKING:
    from .kinds import GenericDict, GenericList, SubsetReturns


@dataclasses.dataclass
class Dictionary(base.Bunch, MutableMapping):
    """Basic bunches dict replacement.

    A `Dictionary` differs from an ordinary python `dict` in ways inherited from
    `Bunch` by requiring `add` and `subset` methods, storing data in `contents`,
    and allowing the "+" operator to join `Dictionary` instances with other
    mappings, including `Dictionary` instances.

    In addition, it differs in 2 other significant ways:
        1) When returning `keys`, `values` and `items`, this class returns them
            as tuples instead of `KeysView`, `ValuesView`, and `ItemsView`.
        2) It includes the same functionality as `defaultdict` in the python
            standard library, including a `setdefault` method.

    Args:
        contents: stored dictionary. Defaults to an empty `dict`.
        default_factory: default value to return or default callable to use to
            create the default value.

    """

    contents: GenericDict = dataclasses.field(default_factory = dict)
    default_factory: Any | None = None

    """ Class Methods """

    @classmethod
    def fromkeys(
        cls,
        keys: GenericList,
        value: Any,
        **kwargs: Any) -> Dictionary:
        """Emulates the `fromkeys` class method from a python `dict`.

        Args:
            keys: items to be keys in a new `Dictionary`.
            value: the value to use for all values in a new `Dictionary`.
            kwargs: additional arguments to pass to the `dict.fromkeys` method.

        Returns:
            An instance formed from `keys` and `value`.

        """
        return cls(contents = dict.fromkeys(keys, value), **kwargs)

    """ Instance Methods """

    def add(self, item: GenericDict, **kwargs: Any) -> None:
        """Adds `item` to the `contents` attribute.

        Args:
            item: items to add to `contents` attribute.
            kwargs: creates a consistent interface even when subclasses have
                additional parameters.

        """
        self.contents.update(item, **kwargs)
        return

    def delete(self, item: Hashable) -> None:
        """Deletes `item` in `contents`.

        Args:
            item: key in `contents` to delete the key/value pair.

        """
        del self.contents[item]
        return

    def get(self, key: Hashable, default: Any | None = None) -> Any:
        """Returns value in `contents` or default options.

        Args:
            key: key for value in `contents`.
            default: default value to return if `key` is not found in
                `contents`.

        Raises:
            KeyError: if `key` is not in `contents` and `default` and the
                `default_factory` attribute are both `None`.

        Returns:
            Value matching key in `contents` or a default value.

        """
        try:
            return self[key]
        except (KeyError, TypeError) as error:
            if default is not None:
                return default
            if self.default_factory is None:
                raise KeyError(f'{key} is not in the Dictionary') from error
            try:
                return self.default_factory()
            except TypeError:
                return self.default_factory

    def items(self) -> tuple[tuple[Hashable, Any], ...]:
        """Emulates python dict `items` method.

        Returns:
            A `tuple` equivalent to `dict.items()`.

        """
        return tuple(zip(self.keys(), self.values(), strict = True))

    def keys(self) -> tuple[Hashable, ...]:
        """Returns `contents` keys as a tuple.

        Returns:
            A `tuple` equivalent to `dict.keys()`.

        """
        return tuple(self.contents.keys())

    def setdefault(self, value: Any) -> None:
        """Sets default value to return when `get` method is used.

        Args:
            value: default value to return when `get` is called and the
                `default` parameter to `get` is None.

        """
        self.default_factory = value
        return

    def subset(
        self,
        include: Collection[Any] | Any | None = None,
        exclude: Collection[Any] | Any | None = None,
        returns: SubsetReturns = settings._SUBSET_RETURN) -> GenericDict:
        """Returns a new instance with a subset of `contents`.

        This method applies `include` before `exclude` if both are passed. If
        `include` is None, all existing items will be added to the new subset
        class instance before `exclude` is applied.

        Args:
            include: item(s) to include in the new `Dictionary`. Defaults to
                `None`.
            exclude: item(s) to exclude from the new `Dictionary`. Defaults to
                `None`.
            returns: whether to return a new instance of the `Dictionary`
                subclass ("class"), a deep copy of the subclass instance
                ("copy") or the simple native Python type ("simple"). Defaults
                to the global setting stored in `settings._SUBSET_RETURN`.

        Raises:
            ValueError: if `include` and `exclude` are both None.

        Returns:
            `dict`-like object with only keys from `include` and no keys in
                `exclude`, in the form dictated by the `returns` argument.

        """
        if include is None and exclude is None:
            raise ValueError('include or exclude must not be None')
        if include is None:
            contents = copy.deepcopy(self.contents)
        else:
            include = list(utilities._iterify(include))
            contents = {k: self.contents[k] for k in include}
        if exclude is not None:
            exclude = list(utilities._iterify(exclude))
            contents = {k: v for k, v in contents.items() if k not in exclude}
        return utilities._return_subset(
            subset = contents,
            existing = self,
            returns = returns)

    def values(self) -> tuple[Any, ...]:
        """Returns `contents` values as a `tuple`.

        Returns:
            A `tuple` equivalent to `dict.values()`.

        """
        return tuple(self.contents.values())

    """ Dunder Methods """

    def __getitem__(self, key: Hashable) -> Any:
        """Returns value for `key` in `contents`.

        Args:
            key: key in `contents` for which a value is sought.

        Returns:
            Value stored in `contents`.

        """
        return self.contents[key]

    def __setitem__(self, key: Hashable, value: Any) -> None:
        """Sets `key` in `contents` to `value`.

        Args:
            key: key to set in `contents`.
            value: value to be paired with `key` in `contents`.

        """
        self.contents[key] = value
        return


@dataclasses.dataclass
class Catalog(Dictionary):
    """Wildcard and list-accepting dictionary.

    A Catalog inherits the differences between a Dictionary and an ordinary
    python dict.

    A Catalog differs from a Dictionary in 5 significant ways:
        1) It recognizes an `all` key which will return a list of all values
            stored in a Catalog instance.
        2) It recognizes a `default` key which will return all values matching
            keys listed in the `default` attribute. `default` can also be set
            using the `catalog['default'] = new_default` assignment. If
            `default` is not passed when the instance is initialized, the
            initial value of `default` is `all`.
        3) It recognizes a `none` key which will return an empty list.
        4) It supports a list of keys being accessed with the matching values
            returned. For example, `catalog[['first_key', 'second_key']]` will
            return the values for those keys in a list ['first_value',
            'second_value'].
        5) If a single key is sought, a Catalog can either return the stored
            value or a stored value in a list (if `always_return_list` is
            True). The latter option is available to make iteration easier
            when the iterator assumes a single type will be returned.

    Args:
        contents: stored dictionary. Defaults to an empty `dict`.
        default_factory: default value to return or default callable to use to
            create the default value.
        default: a list of keys in `contents` which will be used to return items
            when `default` is sought. If not passed, `default` will be set to
            all keys.
        always_return_list: whether to return a list even when the key
            passed is not a list or special access key (`True`) or to return a
            list only when a list or special access key is used (`False`).
            Defaults to `False`.

    """

    contents: GenericDict = dataclasses.field(default_factory = dict)
    default_factory: Any | None = None
    default: Any | None = 'all'
    always_return_list: bool = False

    """ Instance Methods """

    def delete(self, item: Hashable | Sequence[Hashable]) -> None:
        """Deletes `item` in `contents`.

        Args:
            item: name(s) of key(s) in `contents` to delete the key/value pair.

        """
        keys = list(utilities._iterify(item))
        if all(k in self for k in keys):
            self.contents = {
                i: self.contents[i] for i in self.contents if i not in keys}
        else:
            raise KeyError(f'{item} not found in the Catalog')
        return

    """ Dunder Methods """

    def __getitem__(
        self,
        key: Hashable | Sequence[Hashable]) -> Any | GenericList:
        # sourcery skip: assign-if-exp
        """Returns value(s) for `key` in `contents`.

        The method searches for 'all', 'default', and 'none' matching wildcard
        options before searching for direct matches in `contents`.

        Args:
            key: key(s) in `contents`.

        Returns:
            Value(s) stored in `contents`.

        """
        # Returns a list of all values if the 'all' key is sought.
        if key in settings._ALL_KEYS:
            return list(self.contents.values())
        elif key in settings._DEFAULT_KEYS:
            return self[self.default]
        elif key in settings._NONE_KEYS:
            if self.default_factory is None:
                return [] if self.always_return_list else None
            try:
                return self.default_factory()
            except TypeError:
                return self.default_factory
        elif isinstance(key, Sequence) and not isinstance(key, str):
            return [self.contents[k] for k in key if k in self.contents]
        else:
            try:
                if self.always_return_list:
                    return [self.contents[key]]
                else:
                    return self.contents[key]
            except KeyError as error:
                message = f'{key} is not in {self.__class__.__name__}'
                raise KeyError(message) from error

    def __setitem__(
        self,
        key: Hashable | Sequence[Hashable],
        value: Any | GenericList) -> None:
        """Sets `key` in `contents` to `value`.

        Args:
            key: key(s) to set in `contents`.
            value: value(s) to be paired with `key` in `contents`.

        """
        try:
            self.contents[key] = value
        except TypeError:
            self.contents.update(dict(zip(key, value, strict = True)))
        return


@dataclasses.dataclass
class ChainDict(Dictionary):
    """Combines functionality of `collections.ChainMap` with `Dictionary`.

    Args:
        contents: list of stored `Dictionary` instances. This is equivalent to
            the `maps` attribute of a `collections.ChainMap` instance but uses a
            different name for compatibility with `base.Bunch`. A separate
            `maps` property is included which points to `contents` to ensure
            compatibility in the opposite direction.
        default_factory: default value to return or default callable to use to
            create the default value.
        return_first: whether to only return the first match found (`True`) or
            to search all of the stored `Dictionary` instances (`False`).
            Defaults to `True`.

    """

    contents: MutableSequence[Dictionary[Hashable, Any]] = dataclasses.field(
        default_factory = list)
    default_factory: Any | None = None
    return_first: bool | None = True

    """ Properties """

    @property
    def maps(self) -> MutableSequence[Dictionary[Hashable, Any]]:
        """Returns `contents` attribute.

        Returns:
            Stored `Dictionary` instances.

        """
        return self.contents

    @maps.setter
    def maps(self, value: MutableSequence[Dictionary[Hashable, Any]]) -> None:
        """Sets `contents` to `value`.

        Args:
            value: new `list`-like instance to assign `contents` to.

        """
        self.contents = value
        return

    @maps.deleter
    def maps(self) -> None:
        """Sets `contents` to an empty list."""
        self.contents = []
        return

    """ Class Methods """

    @classmethod
    def fromkeys(
        cls,
        keys: Sequence[Hashable],
        value: Any,
        **kwargs: Any) -> Dictionary:
        """Emulates the `fromkeys` class method from a python `dict`.

        Since this method is an awkward fit with a chained map, it just assigns
        the `keys` and `value` to a single `Dictionary` stored in the `contents`
        list.

        Args:
            keys: items to be keys in a new `Dictionary`.
            value: the value to use for all values in a new `Dictionary`.
            kwargs: additional arguments to pass to the `dict.fromkeys` method.

        Returns:
            `Dictionary` formed from `keys` and `value`.

        """
        return cls(contents = [Dictionary.fromkeys(keys, value, **kwargs)])

    """ Instance Methods """

    def add(self, item: GenericDict, **kwargs: Any) -> None:
        """Adds `item` to the `contents` attribute.

        Args:
            item: items to add to `contents` attribute.
            kwargs: creates a consistent interface even when subclasses have
                additional parameters.

        """
        self.contents.append(item, **kwargs)
        return

    def delete(self, item: Hashable) -> None:
        """Deletes `item` in `contents`.

        Because a chained mapping can have identical keys in different stored
        mappings, this method searches through all of the stored Dictionary
        instances and removes the key wherever it appears.

        Args:
            item: key in `contents` to delete the key/value pair.

        """
        for dictionary in self.contents:
            with contextlib.suppress(KeyError):
                del dictionary[item]

        return

    def keys(self) -> tuple[Hashable, ...]:
        """Returns `contents` keys as a `tuple`.

        Returns:
            A `tuple` equivalent to `dict.keys()`.

        """
        return tuple(
            itertools.chain.from_iterable([d.keys() for d in self.contents]))

    def new_child(self, m: Dictionary) -> None:
        """Inserts `m` as the first Dictionary in `contents`.

        This method mirrors the functionality and parameters of
        `collections.Chainmap.new_child`.

        Args:
            m: A new `Dictionary` to add to `contents` at index 0.
        """
        self.contents.insert(0, m)
        return

    def parents(self) -> ChainDict:
        """Returns an instance with `contents` after the first.

        This method mirrors the functionality of `collections.Chainmap.parents`.

        Returns:
            An isntance with all stored `Dictionary` instances after the first.

        """
        return self.__class__(
            self.contents[1:],
            default_factory = self.default_factory)

    def subset(
        self,
        include: Hashable | Sequence[Hashable] | None = None,
        exclude: Hashable | Sequence[Hashable] | None = None,
        returns: SubsetReturns = settings._SUBSET_RETURN) -> ChainDict:
        """Returns a new instance with a subset of `contents`.

        This method applies `include` before `exclude` if both are passed. If
        `include` is None, all existing items will be added to the new subset
        class instance before `exclude` is applied.

        This method relies on all stored mappings being compatible with the
        `Dictionary` class because it uses the `subset` method of those stored
        mappings.

        Args:
            include: item(s) to include in the new `Dictionary`. Defaults to
                `None`.
            exclude: item(s) to exclude from the new `Dictionary`. Defaults to
                `None`.
            returns: whether to return a new instance of the `Dictionary`
                subclass ("class"), a deep copy of the subclass instance
                ("copy") or the simple native Python type ("simple"). Defaults
                to the global setting stored in `settings._SUBSET_RETURN`.

        Raises:
            ValueError: if `include` and `exclude` are both None.

        Returns:
            A new instance ith only keys from `include` and no keys in
                `exclude`.

        """
        if include is None and exclude is None:
            raise ValueError('include or exclude must not be None')
        new_contents = [
            dictionary.subset(include = include, exclude = exclude)
            for dictionary in self.contents]
        return utilities._return_subset(
            subset = new_contents,
            existing = self,
            returns = returns)

    def values(self) -> tuple[Any, ...]:
        """Returns `contents` values as a `tuple`.

        Returns:
            A `tuple` equivalent to `dict.values()`.

        """
        return tuple(
            itertools.chain.from_iterable([d.values() for d in self.contents]))

    """ Dunder Methods """

    def __getitem__(self, key: Hashable) -> Any:
        """Returns value(s) for `key` in `contents`.

        If there are multiple matches for `key` and the `return_first` attribute
        is `False`, this method returns all matches in a `list`. Otherwise, only
        the first match is returned

        Args:
            key: key in `contents` for which a value is sought.

        Returns:
            Value(s) stored in `contents`.

        """
        matches = []
        for dictionary in self.contents:
            with contextlib.suppress(KeyError):
                matches.append(dictionary[key])
                if self.return_first:
                    return matches[0]
        if not matches:
            raise KeyError(f'{key} is not found in the ChainDict')
        return matches[0] if len(matches) > 1 else matches

    def __setitem__(self, key: Hashable, value: Any) -> None:
        """Sets `key` in `contents` to `value`.

        This method stores the passed `key` and `value` in the first stored
        `Dictionary`. If none exists, one is created to stored `key` and
        `value`.

        Args:
            key: key to set in `contents`.
            value: value to be paired with `key` in `contents`.

        """
        if len(self) == 0:
            self.contents = [Dictionary({key: value})]
        else:
            self.contents[0].update({key: value})
        return


@dataclasses.dataclass
class Repository(Dictionary):
    """Dictionary with inferred keys based on items added.

    A `Repository` differs from an ordinary python `dict` in ways inherited from
    `Dictionary`. In addition, it differs in 2 other significant ways:
        1) The `add` method relies on the internal `__get_name__` method to
            assign a str key for the passed item.
        2) It includes an `overwrite` parameter which allows users to determine
            whether existing items will be overwritten when the inferred key
            matches an existing one or whether a new key will be inferred by
            adding an integer counter as a suffix to the key.

    Args:
        contents: stored dictionary. Defaults to an empty `dict`.
        default_factory: default value to return or default callable to use to
            create the default value.
        overwrite: whether to overwrite existing items in the stored dictionary
            with the same inferred keys (`True`) or automatically infer a new
            key based upon a counter suffix (`False`). Defaults to `False`.

    """

    contents: GenericDict = dataclasses.field(default_factory = dict)
    default_factory: Any | None = None
    overwrite: bool | None = False

    """ Instance Methods """

    def add(self, item: Any, key: str | None = None, **kwargs: Any) -> None:
        """Adds `item` to the `contents` attribute.

        Args:
            item: item to add to `contents` attribute.
            key: key to use for `item` if the user does not want the key to be
                inferred.
            kwargs: creates a consistent interface even when subclasses have
                additional parameters.

        """
        key = key or self._get_name(item = item)
        if not self.overwrite:
            key = utilities._uniquify(key = key, dictionary = self)
        self.contents.update({key: item}, **kwargs)
        return

    """ Private Methods """

    def _get_name(self, item: Any) -> str:
        """Infers key name for `item`

        By default, this method uses the `namify` function in bunches. Override
        this method to use a different naming function.

        Args:
            item: item to infer the name for.

        Returns:
            Inferred name.

        """
        return utilities._namify(item)
